# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np

from kolena._utils import log
from kolena.workflow.annotation import Label
from kolena.workflow.annotation import ScoredLabel
from kolena.workflow.metrics import f1_score
from kolena.workflow.metrics import precision
from kolena.workflow.metrics import recall
from kolena.workflow.plot import ConfusionMatrix
from kolena.workflow.plot import Curve
from kolena.workflow.plot import CurvePlot
from kolena.workflow.plot import Histogram


def get_label_confidence(label: str, inference_labels: List[ScoredLabel]) -> float:
    """
    Returns the confidence score of the specified `label` from a list of confidence scores for each label.

    :param label: The label whose confidence score to return.
    :param inference_labels: The list of confidence scores for each label. For `N`-class problem, expected to have
        `N` entries, one for each class.
    :return: The confidence score of the specified `label`. If the `label` doesn't exist in `inference_labels`
        then returns 0.
    """
    return next((inf.score for inf in inference_labels if inf.label == label), 0)


def get_histogram_range(values: List[float]) -> Optional[Tuple[float, float, int]]:
    """
    Computes an ideal range for a confidence score histograms given a list of confidence scores.

    :param values: The list of confidence scores, [0, 1].
    :return: A tuple of min, max and # of bins for a confidence score histograms. The range is rounded up/down to
        the nearest 0.02. The bin size is 0.02.
    """
    if len(values) == 0:
        log.warn("insufficient values provided for confidence histograms")
        return None

    NUM002 = 0.02
    lower = min(values)
    higher = max(values)
    if lower < 0.0 or higher > 1.0:
        log.warn(
            f"values out of range for confidence histograms: expecting [0, 1], got [{lower:.3f}, {higher:.3f}]",
        )
        return None

    # round to 0.02
    min_score = (lower + 1e-9) // NUM002 * NUM002
    max_score = (higher - 1e-9) // NUM002 * NUM002 + NUM002

    if max_score == min_score:
        if max_score < 0.5:
            max_score = min_score + NUM002
        else:
            min_score = max_score - NUM002

    return min_score, max_score, int((max_score - min_score - 1e-9) // NUM002) + 1


def create_histogram(
    values: List[float],
    range: Tuple[float, float, int],
    title: str = "",
    x_label: str = "",
    y_label: str = "",
) -> Histogram:
    """
    Creates a [`Histogram`][kolena.workflow.plot.Histogram] for the specified range and the number of bins.

    :param values: The list of confidence scores to plot.
    :param range: The min, max and # of bins of the histogram.
    :param title: The title of the plot.
    :param x_label: The label on the x-axis.
    :param y_label: The label on the y-axis.
    :return: The [`Histogram`][kolena.workflow.plot.Histogram].
    """
    min_range, max_range, bins = range
    frequency, buckets = np.histogram(
        values,
        bins=bins,
        range=(min_range, max_range),
    )
    return Histogram(
        title=title,
        x_label=x_label,
        y_label=y_label,
        buckets=list(buckets),
        frequency=list(frequency),
    )


def compute_confusion_matrix(
    ground_truths: List[str],
    inferences: List[str],
    title: str = "Confusion Matrix",
    labels: Optional[List[str]] = None,
) -> Optional[ConfusionMatrix]:
    """
    Computes confusion matrix given a list of ground truth and inference labels.

    For a binary classification case, a 2x2 confusion matrix with the count of TP, FP, FN, and TP is computed.

    :param ground_truths: The ground truth labels.
    :param inferences: The inference labels.
    :param title: The title of confusion matrix.
    :param labels: The list of labels to index the matrix. This may be used to reorder or select a subset of labels.
        By default, labels that appear at least once in `ground_truths` or `inferences` are used in sorted order.
    :return: The [`ConfusionMatrix`][kolena.workflow.plot.ConfusionMatrix].

    """
    if len(ground_truths) != len(inferences):
        log.warn(
            f"ground truth labels ({len(ground_truths)}) and inference labels ({len(inferences)}) "
            "differ in length for a confusion matrix — expecting equal length",
        )
        return None

    if len(ground_truths) == 0:
        log.warn("no ground_truths provided for a confusion matrix")
        return None

    confusion_matrix: Dict[str, Dict[str, int]] = defaultdict(lambda: defaultdict(int))

    for gt, inf in zip(ground_truths, inferences):
        confusion_matrix[gt][inf] += 1

    if labels is None:
        labels = list({label for label in ground_truths} | {label for label in inferences})
        labels = sorted(labels)

    if len(labels) < 2:
        log.warn(f"not enough unique labels — expecting at least 2 unique labels, received {len(labels)}")
        return None

    matrix = []
    for actual_label in labels:
        matrix.append([confusion_matrix[actual_label][predicted_label] for predicted_label in labels])
    return ConfusionMatrix(title=title, labels=labels, matrix=matrix)


def _roc_curve(y_true_input: List[int], y_score_input: List[float]) -> Tuple[List[float], List[float]]:
    # Convert inputs to numpy arrays
    y_true = np.array(y_true_input)
    y_score = np.array(y_score_input)
    # Sort the predictions by descending order of confidence
    sorted_indices = np.argsort(y_score)[::-1]
    y_score = y_score[sorted_indices]
    y_true = y_true[sorted_indices]
    distinct_value_indices = np.where(np.diff(y_score[sorted_indices]))[0]
    threshold_idxs = np.r_[distinct_value_indices, y_true.size - 1]
    # Compute the cumulative sums of true positives and false positives
    tps = np.cumsum(y_true)[threshold_idxs]
    fps = np.cumsum(1 - y_true)[threshold_idxs]
    # Drop collinear points (copied from sklearn)
    if len(fps) > 2:
        optimal_indices = np.where(np.r_[True, np.logical_or(np.diff(fps, 2), np.diff(tps, 2)), True])[0]
        fps = fps[optimal_indices]
        tps = tps[optimal_indices]
    fps = np.r_[0, fps]
    tps = np.r_[0, tps]
    # Map cumulative sums to tpr and fpr
    if fps[-1] <= 0:
        # No negative samples in y_true, false positive value should be meaningless
        fpr = []
    else:
        fpr = fps / fps[-1]
        fpr = fpr.tolist()  # type: ignore
    if tps[-1] <= 0:
        # No positive samples in y_true, true positive value should be meaningless
        tpr = []
    else:
        tpr = tps / tps[-1]
        tpr = tpr.tolist()  # type: ignore
    return fpr, tpr


def compute_roc_curves(
    ground_truths: List[Optional[Label]],
    inferences: List[List[ScoredLabel]],
    labels: Optional[List[str]] = None,
    title: Optional[str] = None,
) -> Optional[CurvePlot]:
    """
    Computes OvR (one-vs-rest) ROC (receiver operating characteristic) curves for each class appears in `ground_truths`
    if not specified.

    :param ground_truths: The list of ground truth [`Label`][kolena.workflow.annotation.Label]. For binary
        classification, the negative class can be `None`.
    :param inferences: The list of inference [`ScoredLabel`][kolena.workflow.annotation.ScoredLabel]. For `N`-class
        problems, each inference is expected to contain `N` entries, one for each class and its associated confidence
        score.
    :param labels: The labels to plot. If not specified, classes appear in `ground_truths` are used. Use `labels` to
        specify the evaluating classes especially if `ground_truths` only have negative classes.
    :param title: The title of the plot.
    :return: A [`CurvePlot`][kolena.workflow.plot.CurvePlot] if there is any valid `Curve` computed; otherwise, `None`.
    """
    if len(ground_truths) != len(inferences):
        log.warn(
            f"ground_truths ({len(ground_truths)}) and inferences ({len(inferences)}) differ in length",
        )
        return None

    if len(ground_truths) <= 2:
        log.warn(
            f"insufficient # of samples to compute a roc curve — need at least 3 but received {len(ground_truths)}",
        )
        return None

    if labels is None:
        labels = sorted({gt.label for gt in ground_truths if gt is not None})

    curves: List[Curve] = []
    for label in labels:
        y_true = [1 if gt is not None and gt.label == label else 0 for gt in ground_truths]
        y_score = [get_label_confidence(label, inf) for inf in inferences]
        fpr_values, tpr_values = _roc_curve(y_true_input=y_true, y_score_input=y_score)

        if len(fpr_values) > 0 and len(tpr_values) > 0 and len(fpr_values) == len(tpr_values):
            curves.append(Curve(x=fpr_values, y=tpr_values, label=label))

    inference_labels = [inf.label for inf in inferences[0]]
    is_binary = len(inference_labels) == 1

    if len(curves) > 0:
        if title is None:
            title_prefix = "Receiver Operating Characteristic"
            title = title_prefix if is_binary else title_prefix + " (One-vs-Rest)"
        return CurvePlot(
            title=title,
            x_label="False Positive Rate (FPR)",
            y_label="True Positive Rate (TPR)",
            curves=curves,
        )
    return None


def compute_threshold_curves(
    ground_truths: List[Optional[Label]],
    inferences: List[ScoredLabel],
    thresholds: Optional[List[float]] = None,
) -> Optional[List[Curve]]:
    """
    Computes scores (i.e. [Precision][kolena.workflow.metrics.precision], [Recall][kolena.workflow.metrics.recall]
    and [F1-score][kolena.workflow.metrics.f1_score]) vs. threshold curves for a **single** class presented in
    `inferences`.

    Expects `ground_truths` and `inferences` correspond to the same sample for the same given index.

    :param ground_truths: The list of ground truth [`Label`][kolena.workflow.annotation.Label]s. For binary
        classification, the negative class can be `None`.
    :param inferences: The list of inference [`ScoredLabel`][kolena.workflow.annotation.ScoredLabel]s. The length of
        `inferences` must match the length of `ground_truths`. The list should only include inferences of a specific
        class to plot the threshold curves for.
    :param thresholds: The list of thresholds to plot with. If not specified, all the unique confidence scores are used
        as thresholds, including evenly spaced thresholds from 0 to 1 with 0.1 step.
    :return: A list of [`Curve`][kolena.workflow.plot.Curve]s if there is any valid `Curve` computed; otherwise, `None`.

    """
    if len(ground_truths) != len(inferences):
        log.warn(f"ground_truths ({len(ground_truths)}) and inferences ({len(inferences)}) differ in length")
        return None

    if len(ground_truths) == 0 or len(inferences) == 0:
        log.warn(
            "insufficient # of samples to compute threshold curves — need at least 1 but received "
            f"{len(ground_truths)}",
        )
        return None

    inference_label_list = list({inf.label for inf in inferences})
    if len(inference_label_list) > 1:
        log.warn(
            f"more than one class passed in as inferences: {inference_label_list} — expecting inferences belonging to "
            "a single class",
        )
        return None

    inference_label = inference_label_list[0]
    gts = [gt.label == inference_label if gt else False for gt in ground_truths]

    if thresholds is None:
        unique_scores = list({inf.score for inf in inferences})
        initial_thresholds = list(np.linspace(0, 1, 11))
        thresholds = [0.0]
        for threshold in sorted(set(initial_thresholds + unique_scores)):
            if abs(threshold - thresholds[-1]) >= 1e-2:
                thresholds.append(threshold)
    else:
        thresholds = sorted(set(thresholds))

    precisions = []
    recalls = []
    f1s = []
    for threshold in thresholds:
        infs = [inf.score >= threshold for inf in inferences]
        tp = len([True for gt, inf in zip(gts, infs) if gt and inf])
        fp = len([True for gt, inf in zip(gts, infs) if not gt and inf])
        fn = len([True for gt, inf in zip(gts, infs) if gt and not inf])

        precisions.append(precision(tp, fp))
        recalls.append(recall(tp, fn))
        f1s.append(f1_score(tp, fp, fn))

    return [
        Curve(x=thresholds, y=precisions, label="Precision"),
        Curve(x=thresholds, y=recalls, label="Recall"),
        Curve(x=thresholds, y=f1s, label="F1"),
    ]



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
from collections import defaultdict
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Literal
from typing import Union

import pandas as pd

from kolena import dataset
from kolena._experimental.object_detection.utils import compute_optimal_f1_threshold
from kolena._experimental.object_detection.utils import compute_optimal_f1_threshold_multiclass
from kolena._experimental.object_detection.utils import filter_inferences
from kolena.annotation import ScoredLabel
from kolena.errors import IncorrectUsageError
from kolena.metrics import InferenceMatches
from kolena.metrics import match_inferences
from kolena.metrics import match_inferences_multiclass
from kolena.metrics import MulticlassInferenceMatches


def single_class_datapoint_metrics(object_matches: InferenceMatches, thresholds: float) -> Dict[str, Any]:
    tp = [{**gt._to_dict(), **inf._to_dict()} for gt, inf in object_matches.matched if inf.score >= thresholds]
    fp = [inf for inf in object_matches.unmatched_inf if inf.score >= thresholds]
    fn = object_matches.unmatched_gt + [gt for gt, inf in object_matches.matched if inf.score < thresholds]
    scores = [inf["score"] for inf in tp] + [inf.score for inf in fp]
    return dict(
        TP=tp,
        FP=fp,
        FN=fn,
        matched_inference=[{**gt._to_dict(), **inf._to_dict()} for gt, inf in object_matches.matched],
        unmatched_ground_truth=object_matches.unmatched_gt,
        unmatched_inference=object_matches.unmatched_inf,
        count_TP=len(tp),
        count_FP=len(fp),
        count_FN=len(fn),
        has_TP=len(tp) > 0,
        has_FP=len(fp) > 0,
        has_FN=len(fn) > 0,
        max_confidence_above_t=max(scores) if len(scores) > 0 else None,
        min_confidence_above_t=min(scores) if len(scores) > 0 else None,
        thresholds=thresholds,
    )


def multiclass_datapoint_metrics(
    object_matches: MulticlassInferenceMatches,
    thresholds: Dict[str, float],
) -> Dict[str, Any]:
    tp = [
        {**gt._to_dict(), **inf._to_dict()} for gt, inf in object_matches.matched if inf.score >= thresholds[inf.label]
    ]
    fp = [inf for inf in object_matches.unmatched_inf if inf.score >= thresholds[inf.label]]
    fn = [gt for gt, _ in object_matches.unmatched_gt] + [
        gt for gt, inf in object_matches.matched if inf.score < thresholds[inf.label]
    ]
    unmatched_ground_truth = [
        dict(**gt._to_dict(), predicted_label=inf.label, predicted_score=inf.score) if inf else gt
        for gt, inf in object_matches.unmatched_gt
    ]
    confused = [
        dict(**inf._to_dict(), actual_label=gt.label)
        for gt, inf in object_matches.unmatched_gt
        if inf is not None and inf.score >= thresholds[inf.label]
    ]
    scores = [inf["score"] for inf in tp] + [inf.score for inf in fp]
    inference_labels = {inf.label for _, inf in object_matches.matched}.union(
        {inf.label for inf in object_matches.unmatched_inf},
    )
    fields = [
        ScoredLabel(label=label, score=thresholds[label])
        for label in sorted(thresholds.keys())
        if label in inference_labels
    ]
    return dict(
        TP=tp,
        FP=fp,
        FN=fn,
        matched_inference=[{**gt._to_dict(), **inf._to_dict()} for gt, inf in object_matches.matched],
        unmatched_ground_truth=unmatched_ground_truth,
        unmatched_inference=object_matches.unmatched_inf,
        Confused=confused,
        count_TP=len(tp),
        count_FP=len(fp),
        count_FN=len(fn),
        count_Confused=len(confused),
        has_TP=len(tp) > 0,
        has_FP=len(fp) > 0,
        has_FN=len(fn) > 0,
        has_Confused=len(confused) > 0,
        max_confidence_above_t=max(scores) if len(scores) > 0 else None,
        min_confidence_above_t=min(scores) if len(scores) > 0 else None,
        thresholds=fields,
    )


def _compute_metrics(
    pred_df: pd.DataFrame,
    *,
    ground_truth: str,
    inference: str,
    iou_threshold: float = 0.5,
    threshold_strategy: Union[Literal["F1-Optimal"], float, Dict[str, float]] = 0.5,
    min_confidence_score: float = 0.5,
) -> pd.DataFrame:
    """
    Compute metrics for object detection.

    :param df: Dataframe for model results.
    :param ground_truth: Column name for ground truth object annotations
    :param inference: Column name for inference object annotations
    :param iou_threshold: The [IoU ↗](../../metrics/iou.md) threshold, defaulting to `0.5`.
    :param threshold_strategy: The confidence threshold strategy. It can either be a fixed confidence threshold such
        as `0.5` or `0.75`, or the F1-optimal threshold.
    :param min_confidence_score: The minimum confidence score to consider for the evaluation. This is usually set to
        reduce noise by excluding inferences with low confidence score.
    """
    is_multiclass = _check_multiclass(pred_df[ground_truth], pred_df[inference])
    match_fn = match_inferences_multiclass if is_multiclass else match_inferences

    idx = {name: i for i, name in enumerate(list(pred_df), start=1)}

    all_object_matches: Union[List[MulticlassInferenceMatches], List[InferenceMatches]] = []
    for record in pred_df.itertuples():
        ground_truths = record[idx[ground_truth]]
        inferences = record[idx[inference]]
        all_object_matches.append(
            match_fn(  # type: ignore[arg-type]
                ground_truths,
                filter_inferences(inferences, min_confidence_score),
                mode="pascal",
                iou_threshold=iou_threshold,
            ),
        )

    thresholds: dict[str, float]

    if is_multiclass:
        if isinstance(threshold_strategy, dict):
            thresholds = threshold_strategy
        elif isinstance(threshold_strategy, float):
            thresholds = defaultdict(lambda: threshold_strategy)
        else:
            thresholds = compute_optimal_f1_threshold_multiclass(
                cast(List[MulticlassInferenceMatches], all_object_matches),
            )
        results = [
            multiclass_datapoint_metrics(cast(MulticlassInferenceMatches, matches), thresholds)
            for matches in all_object_matches
        ]
    else:
        if isinstance(threshold_strategy, dict) and threshold_strategy:
            threshold = next(iter(threshold_strategy.values()))
        elif isinstance(threshold_strategy, float):
            threshold = threshold_strategy
        else:
            threshold = compute_optimal_f1_threshold(cast(List[InferenceMatches], all_object_matches))
        results = [
            single_class_datapoint_metrics(cast(InferenceMatches, matches), threshold) for matches in all_object_matches
        ]

    return pd.concat([pd.DataFrame(results), pred_df], axis=1)


def _check_multiclass(ground_truth: pd.Series, inference: pd.Series) -> bool:
    try:
        labels = {x.label for x in itertools.chain.from_iterable(ground_truth)}.union(
            {x.label for x in itertools.chain.from_iterable(inference)},
        )
        return len(labels) >= 2
    except AttributeError:
        return False


def _validate_column_present(df: pd.DataFrame, col: str) -> None:
    if col not in df.columns:
        raise IncorrectUsageError(f"Missing column '{col}'")


def upload_object_detection_results(
    dataset_name: str,
    model_name: str,
    df: pd.DataFrame,
    *,
    ground_truths_field: str = "ground_truths",
    raw_inferences_field: str = "raw_inferences",
    iou_threshold: float = 0.5,
    threshold_strategy: Union[Literal["F1-Optimal"], float, Dict[str, float]] = "F1-Optimal",
    min_confidence_score: float = 0.01,
) -> None:
    """
    Compute metrics and upload results of the model for the dataset.

    Dataframe `df` should include a `locator` column that would match to that of corresponding datapoint. Column
    :inference in the Dataframe `df` should be a list of scored [`BoundingBoxes`][kolena.annotation.BoundingBox].

    :param dataset_name: Dataset name.
    :param model_name: Model name.
    :param df: Dataframe for model results.
    :param ground_truths_field: Field name in datapoint with ground truth bounding boxes,
    defaulting to `"ground_truths"`.
    :param raw_inferences_field: Column in model result DataFrame with raw inference bounding boxes,
    defaulting to `"raw_inferences"`.
    :param iou_threshold: The [IoU ↗](../../metrics/iou.md) threshold, defaulting to `0.5`.
    :param threshold_strategy: The confidence threshold strategy. It can either be a fixed confidence threshold such
        as `0.5` or `0.75`, or `"F1-Optimal"` to find the threshold maximizing F1 score..
    :param min_confidence_score: The minimum confidence score to consider for the evaluation. This is usually set to
        reduce noise by excluding inferences with low confidence score.
    :return:
    """
    eval_config = dict(
        iou_threshold=iou_threshold,
        threshold_strategy=threshold_strategy,
        min_confidence_score=min_confidence_score,
    )
    _validate_column_present(df, raw_inferences_field)

    dataset_df = dataset.download_dataset(dataset_name)
    dataset_df = dataset_df[["locator", ground_truths_field]]
    _validate_column_present(dataset_df, ground_truths_field)

    results_df = _compute_metrics(
        df.merge(dataset_df, on="locator"),
        ground_truth=ground_truths_field,
        inference=raw_inferences_field,
        threshold_strategy=threshold_strategy,
        iou_threshold=iou_threshold,
        min_confidence_score=min_confidence_score,
    )
    results_df.drop(columns=[ground_truths_field], inplace=True)
    dataset.upload_results(dataset_name, model_name, [(eval_config, results_df)])



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from pydantic.dataclasses import dataclass

from kolena.workflow import define_workflow
from kolena.workflow import EvaluatorConfiguration
from kolena.workflow import GroundTruth as BaseGroundTruth
from kolena.workflow import Image
from kolena.workflow import Inference as BaseInference
from kolena.workflow import Metadata
from kolena.workflow import MetricsTestCase
from kolena.workflow import MetricsTestSample
from kolena.workflow import MetricsTestSuite
from kolena.workflow.annotation import LabeledBoundingBox
from kolena.workflow.annotation import LabeledPolygon
from kolena.workflow.annotation import ScoredLabel
from kolena.workflow.annotation import ScoredLabeledBoundingBox
from kolena.workflow.annotation import ScoredLabeledPolygon


@dataclass(frozen=True)
class TestSample(Image):
    """The [`Image`][kolena.workflow.Image] sample type for the pre-built 2D Object Detection workflow."""

    metadata: Metadata = dataclasses.field(default_factory=dict)
    """The optional [`Metadata`][kolena.workflow.Metadata] dictionary."""


@dataclass(frozen=True)
class GroundTruth(BaseGroundTruth):
    """Ground truth type for the pre-built 2D Object Detection workflow."""

    bboxes: List[Union[LabeledBoundingBox, LabeledPolygon]]
    """
    The ground truth [`LabeledBoundingBox`][kolena.workflow.annotation.LabeledBoundingBox]es or
    [`LabeledPolygon`][kolena.workflow.annotation.LabeledPolygon]s associated with an image.
    """

    ignored_bboxes: List[Union[LabeledBoundingBox, LabeledPolygon]] = dataclasses.field(default_factory=list)
    """
    The ground truth [`LabeledBoundingBox`][kolena.workflow.annotation.LabeledBoundingBox]es or
    [`LabeledPolygon`][kolena.workflow.annotation.LabeledPolygon]s to be ignored in evaluation associated with an
    image.
    """

    n_bboxes: int = dataclasses.field(default_factory=lambda: 0)

    def __post_init__(self) -> None:
        object.__setattr__(self, "n_bboxes", len(self.bboxes))


@dataclass(frozen=True)
class Inference(BaseInference):
    """Inference type for the pre-built 2D Object Detection workflow."""

    bboxes: List[Union[ScoredLabeledBoundingBox, ScoredLabeledPolygon]]
    """
    The inference [`ScoredLabeledBoundingBox`][kolena.workflow.annotation.ScoredLabeledBoundingBox]es or
    [`ScoredLabeledPolygon`][kolena.workflow.annotation.ScoredLabeledPolygon]s associated with an image.
    """

    ignored: bool = False
    """
    Whether the image (and its associated inference `bboxes`) should be ignored in evaluating the results of the model.
    """


_, TestCase, TestSuite, Model = define_workflow(
    "Object Detection",
    TestSample,
    GroundTruth,
    Inference,
)


@dataclass(frozen=True)
class TestSampleMetricsSingleClass(MetricsTestSample):
    TP: List[Union[ScoredLabeledBoundingBox, ScoredLabeledPolygon]]
    FP: List[Union[ScoredLabeledBoundingBox, ScoredLabeledPolygon]]
    FN: List[Union[LabeledBoundingBox, LabeledPolygon]]

    count_TP: int
    count_FP: int
    count_FN: int

    has_TP: bool
    has_FP: bool
    has_FN: bool
    ignored: bool

    max_confidence_above_t: Optional[float]
    min_confidence_above_t: Optional[float]
    thresholds: float


@dataclass(frozen=True)
class TestCaseMetricsSingleClass(MetricsTestCase):
    Objects: int
    Inferences: int
    TP: int
    FN: int
    FP: int
    nIgnored: int
    Precision: float
    Recall: float
    F1: float
    AP: float


@dataclass(frozen=True)
class TestSampleMetrics(MetricsTestSample):
    TP: List[Union[ScoredLabeledBoundingBox, ScoredLabeledPolygon]]
    FP: List[Union[ScoredLabeledBoundingBox, ScoredLabeledPolygon]]
    FN: List[Union[LabeledBoundingBox, LabeledPolygon]]
    Confused: List[Union[ScoredLabeledBoundingBox, ScoredLabeledPolygon]]

    count_TP: int
    count_FP: int
    count_FN: int
    count_Confused: int

    has_TP: bool
    has_FP: bool
    has_FN: bool
    has_Confused: bool
    ignored: bool

    max_confidence_above_t: Optional[float]
    min_confidence_above_t: Optional[float]
    thresholds: List[ScoredLabel]


@dataclass(frozen=True)
class ClassMetricsPerTestCase(MetricsTestCase):
    Class: str
    nImages: int
    Threshold: float
    Objects: int
    Inferences: int
    TP: int
    FN: int
    FP: int
    Precision: float
    Recall: float
    F1: float
    AP: float


@dataclass(frozen=True)
class TestCaseMetrics(MetricsTestCase):
    PerClass: List[ClassMetricsPerTestCase]
    Objects: int
    Inferences: int
    TP: int
    FN: int
    FP: int
    nIgnored: int
    macro_Precision: float
    macro_Recall: float
    macro_F1: float
    mean_AP: float
    micro_Precision: float
    micro_Recall: float
    micro_F1: float


@dataclass(frozen=True)
class TestSuiteMetrics(MetricsTestSuite):
    n_images: int
    mean_AP: float
    threshold: Optional[float] = None


@dataclass(frozen=True)
class ThresholdConfiguration(EvaluatorConfiguration):
    """
    Confidence and [IoU ↗](../../metrics/iou.md) threshold configuration for the pre-built 2D Object Detection workflow.
    Specify a confidence and IoU threshold to apply to all classes.
    """

    threshold_strategy: Union[Literal["F1-Optimal"], float] = "F1-Optimal"
    """The confidence threshold strategy. It can either be a fixed confidence threshold such as `0.3` or `0.75`, or
    the F1-optimal threshold by default."""

    iou_threshold: float = 0.5
    """The [IoU ↗](../../metrics/iou.md) threshold, defaulting to `0.5`."""

    min_confidence_score: float = 0.0
    """
    The minimum confidence score to consider for the evaluation. This is usually set to reduce noise by excluding
    inferences with low confidence score.
    """

    multiclass: Optional[bool] = None
    """Optional flag to choose whether the SingleClass or Multiclass evaluator is run.
    If unset, the evaluator is determined based on the number of distinct labels in test data."""

    def display_name(self) -> str:
        evaluator_name = ""
        if self.multiclass is not None:
            evaluator_name = "Multiclass, " if self.multiclass else "SingleClass, "
        return (
            f"{evaluator_name}"
            f"Confidence Threshold: {self.threshold_strategy}, "
            f"IoU: {self.iou_threshold}, "
            f"min confidence ≥ {self.min_confidence_score}"
        )



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum

from pydantic.dataclasses import dataclass


class BatchedLoad:
    class Path(str, Enum):
        INIT_UPLOAD = "/batched-load/upload/init"
        COMPLETE_DOWNLOAD = "/batched-load/download/complete"
        UPLOAD_SIGNED_URL_STUB = "/batched-load/upload/signed-url"
        DOWNLOAD_BY_PATH_STUB = "/batched-load/download/by-path"

        @classmethod
        def upload_signed_url(cls, load_uuid: str) -> str:
            return f"{cls.UPLOAD_SIGNED_URL_STUB.value}/{load_uuid}"

        @classmethod
        def download_by_path(cls, path: str) -> str:
            return f"{cls.DOWNLOAD_BY_PATH_STUB.value}/{path}"

    @dataclass(frozen=True)
    class WithLoadUUID:
        uuid: str

    @dataclass(frozen=True)
    class SignedURL:
        signed_url: str

    @dataclass(frozen=True)
    class BaseInitDownloadRequest:
        batch_size: int

    @dataclass(frozen=True)
    class InitDownloadPartialResponse(WithLoadUUID):
        path: str

    @dataclass(frozen=True)
    class CompleteDownloadRequest(WithLoadUUID):
        ...

    @dataclass(frozen=True)
    class InitiateUploadResponse(WithLoadUUID):
        ...
